/* This code is part of Freenet. It is distributed under the GNU General
 * Public License, version 2 (or at your option any later version). See
 * http://www.gnu.org/ for further details of the GPL. */
package freenet.clients.http;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.net.URI;
import java.net.URISyntaxException;
import java.security.MessageDigest;
import java.util.Iterator;
import java.util.LinkedList;

import freenet.clients.http.PageMaker.THEME;
import freenet.clients.http.bookmark.BookmarkManager;
import freenet.config.EnumerableOptionCallback;
import freenet.config.InvalidConfigValueException;
import freenet.config.NodeNeedRestartException;
import freenet.config.SubConfig;
import freenet.crypt.SHA256;
import freenet.crypt.SSL;
import freenet.io.AllowedHosts;
import freenet.io.NetworkInterface;
import freenet.io.SSLNetworkInterface;
import freenet.keys.FreenetURI;
import freenet.l10n.L10n;
import freenet.node.Node;
import freenet.node.NodeClientCore;
import freenet.support.Base64;
import freenet.support.Executor;
import freenet.support.HTMLNode;
import freenet.support.Logger;
import freenet.support.OOMHandler;
import freenet.support.URLDecoder;
import freenet.support.URLEncodedFormatException;
import freenet.support.URLEncoder;
import freenet.support.api.BooleanCallback;
import freenet.support.api.BucketFactory;
import freenet.support.api.IntCallback;
import freenet.support.api.LongCallback;
import freenet.support.api.StringCallback;
import freenet.support.io.ArrayBucketFactory;
import freenet.support.io.Closer;

/** 
 * The Toadlet (HTTP) Server
 * 
 * Provide a HTTP server for FProxy
 */
public final class SimpleToadletServer implements ToadletContainer, Runnable, LinkFixer {
	/** List of urlPrefix / Toadlet */ 
	private final LinkedList<ToadletElement> toadlets;
	private static class ToadletElement {
		public ToadletElement(Toadlet t2, String urlPrefix) {
			t = t2;
			prefix = urlPrefix;
		}
		Toadlet t;
		String prefix;
	}

	// Socket / Binding
	private final int port;
	private String bindTo;
	private String allowedHosts;
	private NetworkInterface networkInterface;
	private boolean ssl = false;
	public static final int DEFAULT_FPROXY_PORT = 8888;
	private byte[] clientNonce;
	private final Node node;
	
	// ACL
	private final AllowedHosts allowedFullAccess;
	private boolean publicGatewayMode;
	
	// Theme 
	private THEME cssTheme;
	private File cssOverride;
	private boolean advancedModeEnabled;
	private final PageMaker pageMaker;
	
	// Control
	private Thread myThread;
	private final Executor executor;
	private BucketFactory bf;
	private NodeClientCore core;
	
	// HTTP Option
	private boolean doRobots;
	private boolean enablePersistentConnections;
	private boolean enableInlinePrefetch;
	private boolean enableHistoryCloaking;
	
	// Something does not really belongs to here
	static boolean isPanicButtonToBeShown;				// move to QueueToadlet ?
	public BookmarkManager bookmarkManager;				// move to WelcomeToadlet / BookmarkEditorToadlet ?
	private volatile boolean fProxyJavascriptEnabled;	// ugh?
	private volatile boolean fproxyHasCompletedWizard;	// hmmm..
	
	// Config Callbacks
	private class FProxySSLCallback extends BooleanCallback  {
		@Override
		public Boolean get() {
			return ssl;
		}
		@Override
		public void set(Boolean val) throws InvalidConfigValueException {
			if (get().equals(val))
				return;
			if(!SSL.available()) {
				throw new InvalidConfigValueException("Enable SSL support before use ssl with Fproxy");
			}
			ssl = val;
			throw new InvalidConfigValueException("Cannot change SSL on the fly, please restart freenet");
		}
		@Override
		public boolean isReadOnly() {
			return true;
		}
	}
	
	private static class FProxyPassthruMaxSize extends LongCallback {
		@Override
		public Long get() {
			return FProxyToadlet.MAX_LENGTH;
		}
		
		@Override
		public void set(Long val) throws InvalidConfigValueException {
			if (get().equals(val))
				return;
			FProxyToadlet.MAX_LENGTH = val;
		}
	}

	private class FProxyPortCallback extends IntCallback  {
		private Integer savedPort;
		@Override
		public Integer get() {
			if (savedPort == null)
				savedPort = port;
			return savedPort;
		}
		
		@Override
		public void set(Integer newPort) throws NodeNeedRestartException {
			if(savedPort != (int)newPort) {
				savedPort = port;
				throw new NodeNeedRestartException("Port cannot change on the fly");
			}
		}
	}
	
	private class FProxyBindtoCallback extends StringCallback  {
		@Override
		public String get() {
			return bindTo;
		}
		
		@Override
		public void set(String bindTo) throws InvalidConfigValueException {
			if(!bindTo.equals(get())) {
				try {
					networkInterface.setBindTo(bindTo, false);
					SimpleToadletServer.this.bindTo = bindTo;
				} catch (IOException e) {
					// This is an advanced option for reasons of reducing clutter,
					// but it is expected to be used by regular users, not devs.
					// So we translate the error messages.
					throw new InvalidConfigValueException(l10n("couldNotChangeBindTo", "error", e.getLocalizedMessage()));
				}
			}
		}
	}
	private class FProxyAllowedHostsCallback extends StringCallback  {
		@Override
		public String get() {
			return networkInterface.getAllowedHosts();
		}
		
		@Override
		public void set(String allowedHosts) {
			if (!allowedHosts.equals(get())) {
				networkInterface.setAllowedHosts(allowedHosts);
			}
		}
	}
	private class FProxyCSSNameCallback extends StringCallback implements EnumerableOptionCallback {
		@Override
		public String get() {
			return cssTheme.code;
		}
		
		@Override
		public void set(String CSSName) throws InvalidConfigValueException {
			if((CSSName.indexOf(':') != -1) || (CSSName.indexOf('/') != -1))
				throw new InvalidConfigValueException(l10n("illegalCSSName"));
			cssTheme = THEME.themeFromName(CSSName);
			pageMaker.setTheme(cssTheme);
			if (core.node.pluginManager != null)
				core.node.pluginManager.setFProxyTheme(cssTheme);
		}

		public String[] getPossibleValues() {
			return THEME.possibleValues;
		}
	}
	private class FProxyCSSOverrideCallback extends StringCallback  {
		@Override
		public String get() {
			return (cssOverride == null ? "" : cssOverride.toString());
		}

		@Override
		public void set(String val) throws InvalidConfigValueException {
			if(core == null) return;
			if(val.equals(get()) || val.equals(""))
				cssOverride = null;
			else {
				File tmp = new File(val.trim());
				if(!core.allowUploadFrom(tmp))
					throw new InvalidConfigValueException(l10n("cssOverrideNotInUploads", "filename", tmp.toString()));
				else if(!tmp.canRead() || !tmp.isFile())
					throw new InvalidConfigValueException(l10n("cssOverrideCantRead", "filename", tmp.toString()));
				cssOverride = tmp.getAbsoluteFile();
			}
			pageMaker.setOverride(cssOverride);
		}
	}
	private class FProxyEnabledCallback extends BooleanCallback  {
		@Override
		public Boolean get() {
			synchronized(SimpleToadletServer.this) {
				return myThread != null;
			}
		}
		@Override
		public void set(Boolean val) throws InvalidConfigValueException {
			if (get().equals(val))
				return;
			synchronized(SimpleToadletServer.this) {
				if(val) {
					// Start it
					myThread = new Thread(SimpleToadletServer.this, "SimpleToadletServer");
				} else {
					myThread.interrupt();
					myThread = null;
					return;
				}
			}
			createFproxy();
			myThread.setDaemon(true);
			myThread.start();
		}
	}
	private static class FProxyAdvancedModeEnabledCallback extends BooleanCallback  {
		private final SimpleToadletServer ts;
		
		FProxyAdvancedModeEnabledCallback(SimpleToadletServer ts){
			this.ts = ts;
		}
		
		@Override
		public Boolean get() {
			return ts.isAdvancedModeEnabled();
		}
		
		@Override
		public void set(Boolean val) throws InvalidConfigValueException {
			if (get().equals(val))
				return;
				ts.enableAdvancedMode(val);
		}
	}
	private static class FProxyJavascriptEnabledCallback extends BooleanCallback  {
		
		private final SimpleToadletServer ts;
		
		FProxyJavascriptEnabledCallback(SimpleToadletServer ts){
			this.ts = ts;
		}
		
		@Override
		public Boolean get() {
			return ts.isFProxyJavascriptEnabled();
		}
		
		@Override
		public void set(Boolean val) throws InvalidConfigValueException {
			if (get().equals(val))
				return;
				ts.enableFProxyJavascript(val);
		}
	}
	
	private static class FProxyHistoryCloakingCallback extends BooleanCallback {
		
		FProxyHistoryCloakingCallback(SimpleToadletServer ts) {
			this.ts = ts;
		}
		
		private final SimpleToadletServer ts;

		@Override
		public Boolean get() {
			return ts.enableHistoryCloaking;
		}

		@Override
		public void set(Boolean val) throws InvalidConfigValueException, NodeNeedRestartException {
			boolean enable = val;
			if(enable == ts.enableHistoryCloaking) return;
			this.ts.enableHistoryCloaking = enable;
			this.ts.writeURLFile();
		}
		
	}
	
	private boolean haveCalledFProxy = false;
	
	public void createFproxy() {
		synchronized(this) {
			if(haveCalledFProxy) return;
			haveCalledFProxy = true;
		}
		bookmarkManager = new BookmarkManager(core);
		try {
			FProxyToadlet.maybeCreateFProxyEtc(core, core.node, core.node.config, this, bookmarkManager);
		} catch (IOException e) {
			Logger.error(this, "Could not start fproxy: "+e, e);
			System.err.println("Could not start fproxy:");
			e.printStackTrace();
		}
	}
	

	
	public void writeURLFile() {
		String url = "http://127.0.0.1:"+port;
		if(enableHistoryCloaking)
			url += fixLink("/");
		else
			url += "/";
		File f = new File(node.getNodeDir(), "freenet.url.dat");
		FileOutputStream fos = null;
		try {
			fos = new FileOutputStream(f);
			OutputStreamWriter osw = new OutputStreamWriter(fos); // System charset better in this instance???
			osw.write(url+(File.separatorChar == '\\' ? "\r" : "") + "\n");
			osw.flush();
		} catch (IOException e) {
			Logger.error(this, "Unable to write URL "+url+" to freenet.url.dat: browse.sh/browse.cmd may have difficulty! : "+e, e);
			System.err.println("Unable to write URL "+url+" to freenet.url.dat: browse.sh/browse.cmd may have difficulty! : "+e);
		} finally {
			Closer.close(fos);
		}
	}



	public synchronized void setCore(NodeClientCore core) {
		this.core = core;
	}
	
	/**
	 * Create a SimpleToadletServer, using the settings from the SubConfig (the fproxy.*
	 * config).
	 */
	public SimpleToadletServer(SubConfig fproxyConfig, BucketFactory bucketFactory, Executor executor, Node node) throws IOException, InvalidConfigValueException {

		this.executor = executor;
		this.node = node;
		
		int configItemOrder = 0;
		
		fproxyConfig.register("enabled", true, configItemOrder++, true, true, "SimpleToadletServer.enabled", "SimpleToadletServer.enabledLong",
				new FProxyEnabledCallback());
		
		boolean enabled = fproxyConfig.getBoolean("enabled");
		
		fproxyConfig.register("ssl", false, configItemOrder++, true, true, "SimpleToadletServer.ssl", "SimpleToadletServer.sslLong",
				new FProxySSLCallback());
		fproxyConfig.register("port", DEFAULT_FPROXY_PORT, configItemOrder++, true, true, "SimpleToadletServer.port", "SimpleToadletServer.portLong",
				new FProxyPortCallback());
		fproxyConfig.register("bindTo", NetworkInterface.DEFAULT_BIND_TO, configItemOrder++, true, true, "SimpleToadletServer.bindTo", "SimpleToadletServer.bindToLong",
				new FProxyBindtoCallback());
		fproxyConfig.register("css", "clean", configItemOrder++, false, false, "SimpleToadletServer.cssName", "SimpleToadletServer.cssNameLong",
				new FProxyCSSNameCallback());
		fproxyConfig.register("CSSOverride", "", configItemOrder++, true, false, "SimpleToadletServer.cssOverride", "SimpleToadletServer.cssOverrideLong",
				new FProxyCSSOverrideCallback());
		fproxyConfig.register("advancedModeEnabled", false, configItemOrder++, true, false, "SimpleToadletServer.advancedMode", "SimpleToadletServer.advancedModeLong",
				new FProxyAdvancedModeEnabledCallback(this));
		fproxyConfig.register("javascriptEnabled", false, configItemOrder++, true, false, "SimpleToadletServer.enableJS", "SimpleToadletServer.enableJSLong",
				new FProxyJavascriptEnabledCallback(this));
		fproxyConfig.register("hasCompletedWizard", false, configItemOrder++, true, false, "SimpleToadletServer.hasCompletedWizard", "SimpleToadletServer.hasCompletedWizardLong",
				new BooleanCallback() {
					@Override
					public Boolean get() {
						return fproxyHasCompletedWizard;
					}

					@Override
					public void set(Boolean val) throws InvalidConfigValueException, NodeNeedRestartException {
						if(get().equals(val)) return;
						fproxyHasCompletedWizard = val;
					}
		});
		fproxyHasCompletedWizard = fproxyConfig.getBoolean("hasCompletedWizard");
		
		fproxyConfig.register("showPanicButton", false, configItemOrder++, true, true, "SimpleToadletServer.panicButton", "SimpleToadletServer.panicButtonLong",
				new BooleanCallback(){
				@Override
				public Boolean get() {
					return SimpleToadletServer.isPanicButtonToBeShown;
				}
			
				@Override
				public void set(Boolean value) {
					if(value == SimpleToadletServer.isPanicButtonToBeShown) return;
					else	SimpleToadletServer.isPanicButtonToBeShown = value;
				}
		});
		
		fproxyConfig.register("publicGatewayMode", false, configItemOrder++, true, true, "SimpleToadletServer.publicGatewayMode", "SimpleToadletServer.publicGatewayModeLong", new BooleanCallback() {

			@Override
			public Boolean get() {
				return publicGatewayMode;
			}

			@Override
			public void set(Boolean val) throws InvalidConfigValueException, NodeNeedRestartException {
				publicGatewayMode = val;
			}
			
		});
		publicGatewayMode = fproxyConfig.getBoolean("publicGatewayMode");
		
		// This is OFF BY DEFAULT because for example firefox has a limit of 2 persistent 
		// connections per server, but 8 non-persistent connections per server. We need 8 conns
		// more than we need the efficiency gain of reusing connections - especially on first
		// install.
		
		fproxyConfig.register("enablePersistentConnections", false, configItemOrder++, true, false, "SimpleToadletServer.enablePersistentConnections", "SimpleToadletServer.enablePersistentConnectionsLong",
				new BooleanCallback() {

					@Override
					public Boolean get() {
						synchronized(SimpleToadletServer.this) {
							return enablePersistentConnections;
						}
					}

					@Override
					public void set(Boolean val) throws InvalidConfigValueException {
						synchronized(SimpleToadletServer.this) {
							enablePersistentConnections = val;
						}
					}
		});
		enablePersistentConnections = fproxyConfig.getBoolean("enablePersistentConnections");
		
		// Off by default.
		// I had hoped it would yield a significant performance boost to bootstrap performance
		// on browsers with low numbers of simultaneous connections. Unfortunately the bottleneck
		// appears to be that the node does very few local requests compared to external requests
		// (for anonymity's sake).
		
		fproxyConfig.register("enableInlinePrefetch", false, configItemOrder++, true, false, "SimpleToadletServer.enableInlinePrefetch", "SimpleToadletServer.enableInlinePrefetchLong",
				new BooleanCallback() {

					@Override
					public Boolean get() {
						synchronized(SimpleToadletServer.this) {
							return enableInlinePrefetch;
						}
					}

					@Override
					public void set(Boolean val) throws InvalidConfigValueException {
						synchronized(SimpleToadletServer.this) {
							enableInlinePrefetch = val;
						}
					}
		});
		enableInlinePrefetch = fproxyConfig.getBoolean("enableInlinePrefetch");
		
		// Off by default, installer turns it on.
		fproxyConfig.register("enableHistoryCloaking", false, configItemOrder++, false, false, "SimpleToadletServer.enableHistoryCloaking", "SimpleToadletServer.enableHistoryCloakingLong", new FProxyHistoryCloakingCallback(this));
		enableHistoryCloaking = fproxyConfig.getBoolean("enableHistoryCloaking");
		
		fproxyConfig.register("passthroughMaxSize", 2L*1024*1024, configItemOrder++, true, false, "SimpleToadletServer.passthroughMaxSize", "SimpleToadletServer.passthroughMaxSizeLong", new FProxyPassthruMaxSize());
		FProxyToadlet.MAX_LENGTH = fproxyConfig.getLong("passthroughMaxSize");
		
		fproxyConfig.register("allowedHosts", "127.0.0.1,0:0:0:0:0:0:0:1", configItemOrder++, true, true, "SimpleToadletServer.allowedHosts", "SimpleToadletServer.allowedHostsLong",
				new FProxyAllowedHostsCallback());
		fproxyConfig.register("allowedHostsFullAccess", "127.0.0.1,0:0:0:0:0:0:0:1", configItemOrder++, true, true, "SimpleToadletServer.allowedFullAccess", 
				"SimpleToadletServer.allowedFullAccessLong",
				new StringCallback() {

					@Override
					public String get() {
						return allowedFullAccess.getAllowedHosts();
					}

					@Override
					public void set(String val) throws InvalidConfigValueException {
						allowedFullAccess.setAllowedHosts(val);
					}
			
		});
		allowedFullAccess = new AllowedHosts(fproxyConfig.getString("allowedHostsFullAccess"));
		fproxyConfig.register("doRobots", false, configItemOrder++, true, false, "SimpleToadletServer.doRobots", "SimpleToadletServer.doRobotsLong",
				new BooleanCallback() {
					@Override
					public Boolean get() {
						return doRobots;
					}
					@Override
					public void set(Boolean val) throws InvalidConfigValueException {
						doRobots = val;
					}
		});
		doRobots = fproxyConfig.getBoolean("doRobots");
		
		SimpleToadletServer.isPanicButtonToBeShown = fproxyConfig.getBoolean("showPanicButton");
		this.bf = bucketFactory;
		port = fproxyConfig.getInt("port");
		bindTo = fproxyConfig.getString("bindTo");
		String cssName = fproxyConfig.getString("css");
		if((cssName.indexOf(':') != -1) || (cssName.indexOf('/') != -1))
			throw new InvalidConfigValueException("CSS name must not contain slashes or colons!");
		cssTheme = THEME.themeFromName(cssName);
		pageMaker = new PageMaker(cssTheme, this);
	
		if(!fproxyConfig.getOption("CSSOverride").isDefault()) {
			cssOverride = new File(fproxyConfig.getString("CSSOverride"));
			pageMaker.setOverride(cssOverride);
		} else
			cssOverride = null;
		
		this.advancedModeEnabled = fproxyConfig.getBoolean("advancedModeEnabled");
		toadlets = new LinkedList<ToadletElement>();

		if(SSL.available()) {
			ssl = fproxyConfig.getBoolean("ssl");
		}
		
		this.allowedHosts=fproxyConfig.getString("allowedHosts");

		if(!enabled) {
			Logger.normal(SimpleToadletServer.this, "Not starting FProxy as it's disabled");
			System.out.println("Not starting FProxy as it's disabled");
		} else {
			maybeGetNetworkInterface();
			myThread = new Thread(this, "SimpleToadletServer");
			myThread.setDaemon(true);
		}
		
		// Register static toadlet and startup toadlet
		
		StaticToadlet statictoadlet = new StaticToadlet();
		register(statictoadlet, "/static/", false, false);
		
		// "Freenet is starting up..." page, to be removed at #removeStartupToadlet()
		startupToadlet = new StartupToadlet(statictoadlet);
		register(startupToadlet, "/", false, false);
	}
	
	public StartupToadlet startupToadlet;
	
	public void removeStartupToadlet() {
		unregister(startupToadlet);
		// Ready to be GCed
		startupToadlet = null;
		// Not in the navbar.
	}
	
	private void maybeGetNetworkInterface() throws IOException {
		if (this.networkInterface!=null) return;
		if(ssl) {
			this.networkInterface = SSLNetworkInterface.create(port, this.bindTo, allowedHosts, executor, true);
		} else {
			this.networkInterface = NetworkInterface.create(port, this.bindTo, allowedHosts, executor, true);
		}
	}		

	public boolean doRobots() {
		return doRobots;
	}
	
	public boolean publicGatewayMode() {
		return publicGatewayMode;
	}
	
	public void start() {
		if(myThread != null) try {
			writeURLFile();
			maybeGetNetworkInterface();
			myThread.start();
			Logger.normal(this, "Starting FProxy on "+bindTo+ ':' +port);
			System.out.println("Starting FProxy on "+bindTo+ ':' +port);
		} catch (IOException e) {
			Logger.error(this, "Could not bind network port for FProxy?", e);
		}
	}
	
	public void register(Toadlet t, String urlPrefix, boolean atFront, boolean fullOnly) {
		register(t, urlPrefix, atFront, null, null, fullOnly, null);
	}
	
	public void register(Toadlet t, String urlPrefix, boolean atFront, String name, String title, boolean fullOnly, LinkEnabledCallback cb) {
		ToadletElement te = new ToadletElement(t, urlPrefix);
		if(atFront) toadlets.addFirst(te);
		else toadlets.addLast(te);
		t.container = this;
		if (name != null) {
			pageMaker.addNavigationLink(urlPrefix, name, title, fullOnly, cb);
		}
	}

	public synchronized void unregister(Toadlet t) {
		for(Iterator<ToadletElement> i=toadlets.iterator();i.hasNext();) {
			ToadletElement e = i.next();
			if(e.t == t) {
				i.remove();
				return;
			}
		}
	}
	
	public StartupToadlet getStartupToadlet() {
		return startupToadlet;
	}
	
	public Toadlet findToadlet(URI uri) throws PermanentRedirectException {
		String path = uri.getPath();

		// Show the wizard until dismissed by the user (See bug #2624)
		if(!fproxyHasCompletedWizard) {
			if(!(core != null && core.node != null && (core.node.isOpennetEnabled() || core.node.getPeerNodes().length > 0))) {
				
			if(!(path.startsWith(FirstTimeWizardToadlet.TOADLET_URL) ||
				path.startsWith(StaticToadlet.ROOT_URL)))
				try {
					throw new PermanentRedirectException(new URI(FirstTimeWizardToadlet.TOADLET_URL));
				} catch(URISyntaxException e) { throw new Error(e); }
				
			}
		}

		Iterator<ToadletElement> i = toadlets.iterator();
		while(i.hasNext()) {
			ToadletElement te = i.next();
						
			if(path.startsWith(te.prefix))
					return te.t;
			if(te.prefix.length() > 0 && te.prefix.charAt(te.prefix.length()-1) == '/') {
				if(path.equals(te.prefix.substring(0, te.prefix.length()-1))) {
					URI newURI;
					try {
						newURI = new URI(te.prefix);
					} catch (URISyntaxException e) {
						throw new Error(e);
					}
					throw new PermanentRedirectException(newURI);
				}
			}
		}
		return null;
	}

	public void run() {
		try {
			networkInterface.setSoTimeout(500);
		} catch (SocketException e1) {
			Logger.error(this, "Could not set so-timeout to 500ms; on-the-fly disabling of the interface will not work");
		}
		while(true) {
			synchronized(this) {
				if(myThread == null) return;
			}
			try {
				Socket conn = networkInterface.accept();
				if(Logger.shouldLog(Logger.MINOR, this))
					Logger.minor(this, "Accepted connection");
				SocketHandler sh = new SocketHandler(conn);
				sh.start();
			} catch (SocketTimeoutException e) {
				// Go around again, this introduced to avoid blocking forever when told to quit
				/* FIXME: WTF? Why does networkInterface.accept() not support being interrupted ? */
			} 
		}
	}
	
	public class SocketHandler implements Runnable {

		Socket sock;
		
		public SocketHandler(Socket conn) {
			this.sock = conn;
		}

		void start() {
			executor.execute(this, "HTTP socket handler@"+hashCode());
		}
		
		public void run() {
		    freenet.support.Logger.OSThread.logPID(this);
			boolean logMINOR = Logger.shouldLog(Logger.MINOR, this);
			if(logMINOR) Logger.minor(this, "Handling connection");
			try {
				ToadletContextImpl.handle(sock, SimpleToadletServer.this, pageMaker);
			} catch (OutOfMemoryError e) {
				OOMHandler.handleOOM(e);
				System.err.println("SimpleToadletServer request above failed.");
				Logger.error(this, "OOM in SocketHandler");
			} catch (Throwable t) {
				System.err.println("Caught in SimpleToadletServer: "+t);
				t.printStackTrace();
				Logger.error(this, "Caught in SimpleToadletServer: "+t, t);
			}
			if(logMINOR) Logger.minor(this, "Handled connection");
		}

	}

	public THEME getTheme() {
		return this.cssTheme;
	}

	public void setCSSName(THEME theme) {
		this.cssTheme = theme;
	}

	public synchronized boolean isAdvancedModeEnabled() {
		return this.advancedModeEnabled;
	}
	
	public synchronized void enableAdvancedMode(boolean b){
		advancedModeEnabled = b;
	}

	public synchronized boolean isFProxyJavascriptEnabled() {
		return this.fProxyJavascriptEnabled;
	}
	
	public synchronized void enableFProxyJavascript(boolean b){
		fProxyJavascriptEnabled = b;
	}

	public String getFormPassword() {
		if(core == null) return "";
		return core.formPassword;
	}

	public boolean isAllowedFullAccess(InetAddress remoteAddr) {
		return this.allowedFullAccess.allowed(remoteAddr);
	}
	
	private static String l10n(String key, String pattern, String value) {
		return L10n.getString("SimpleToadletServer."+key, pattern, value);
	}

	private static String l10n(String key) {
		return L10n.getString("SimpleToadletServer."+key);
	}

	public HTMLNode addFormChild(HTMLNode parentNode, String target, String id) {
		HTMLNode formNode =
			parentNode.addChild("form", new String[] { "action", "method", "enctype", "id",  "accept-charset" }, 
					new String[] { target, "post", "multipart/form-data", id, "utf-8"} ).addChild("div");
		formNode.addChild("input", new String[] { "type", "name", "value" }, 
				new String[] { "hidden", "formPassword", getFormPassword() });
		
		return formNode;
	}

	public void setBucketFactory(BucketFactory tempBucketFactory) {
		this.bf = tempBucketFactory;
	}

	public boolean isEnabled() {
		return myThread != null;
	}

	public BookmarkManager getBookmarks() {
		return bookmarkManager;
	}

	public FreenetURI[] getBookmarkURIs() {
		if(bookmarkManager == null) return new FreenetURI[0];
		return bookmarkManager.getBookmarkURIs();
	}

	public boolean enablePersistentConnections() {
		return enablePersistentConnections;
	}

	public boolean enableInlinePrefetch() {
		return enableInlinePrefetch;
	}

	public synchronized boolean allowPosts() {
		return !(bf instanceof ArrayBucketFactory);
	}

	public synchronized BucketFactory getBucketFactory() {
		return bf;
	}

	/** How long should the secureid= be? We don't want the URLs to be really long...
	 * One attempt requires the attacker to modify the DOM, trigger a relayout, and then
	 * read the link color out of the DOM. Relayout can be batched, but if it's too big
	 * it gets really slow and the user notices... The lookup is done to determine whether
	 * to display this element. A miss results in no relayout. Maybe this could be 
	 * optimised down to hundreds of cycles ... in practice it's probably a lot more than 
	 * that. Hopefully by the time probing 96 bits is feasible, browsers will turn off 
	 * history probing by default! */
	static final int SID_LENGTH_BYTES = 12;
	
	public String generateSID(String realPath) throws URLEncodedFormatException {
		MessageDigest md = SHA256.getMessageDigest();
		realPath = prepareForSID(realPath);
		try {
			md.update(realPath.getBytes("UTF-8"));
		} catch (UnsupportedEncodingException e) {
			throw new Error(e);
		}
		md.update(clientNonce);
		byte[] output = md.digest();
		byte[] finalOutput = new byte[SID_LENGTH_BYTES];
		System.arraycopy(output, 0, finalOutput, 0, SID_LENGTH_BYTES);
		SHA256.returnMessageDigest(md);
		return Base64.encode(finalOutput);
	}

	private String prepareForSID(String realPath) throws URLEncodedFormatException {
		StringBuffer sb = new StringBuffer(realPath.length());
		int idx = realPath.indexOf('?');
		String query = null;
		if(idx > -1) {
			query = realPath.substring(idx+1);
			realPath = realPath.substring(0, idx);
		}
		String[] split = realPath.split("/");
		boolean first = true;
		for(String component : split) {
			if(!first) sb.append('/');
			first = false;
			if(component.indexOf('%') > -1)
				component = URLDecoder.decode(component, true);
			component = URLEncoder.minimalEncode(component, "/?");
			sb.append(component);
		}
		if(query != null) {
			sb.append('?');
			split = query.split("&");
			first = true;
			for(String component : split) {
				if(!first) sb.append('&');
				first = false;
				if(component.indexOf('%') > -1)
					component = URLDecoder.decode(component, true);
				component = URLEncoder.minimalEncode(component, "&");
				sb.append(component);
			}
		}
		return sb.toString();
	}



	public boolean isSecureIDCheckingDisabled() {
		return !enableHistoryCloaking;
	}
	
	/* (non-Javadoc)
	 * @see freenet.clients.http.LinkFixer#fixLink(java.lang.String)
	 */
	public String fixLink(String orig) {
		if(isSecureIDCheckingDisabled())
			return orig;
		if((orig.indexOf("?secureid=") > -1) ||
				(orig.indexOf("?") > -1 && orig.substring(orig.indexOf("?")).indexOf("&secureid=") > -1)) {
			Logger.error(this, "Already has a secureid: "+orig);
			return orig;
		}
		String toSign = orig;
		String frag = "";
		int hashIndex = toSign.indexOf('#');
		if(hashIndex != -1) {
			frag = toSign.substring(hashIndex);
			toSign = toSign.substring(0, hashIndex);
		}
		try {
		if(orig.indexOf('?') == -1) {
			return toSign + "?secureid=" + generateSID(toSign) + frag;
		} else {
			return toSign + "&secureid=" + generateSID(toSign) + frag;
		}
		} catch (URLEncodedFormatException e) {
			Logger.error(this, "UNABLE TO ENCODE INTERNALLY GENERATED STRING: "+orig+" because URL decode failure: "+e);
			return orig;
		}
	}
	
	public URI fixLink(URI uri) throws URISyntaxException {
		if(isSecureIDCheckingDisabled())
			return uri;
		String toSign = uri.toString();
		if(uri.getFragment() != null) {
			toSign = toSign.substring(0, toSign.indexOf('#'));
		}
		try {
		if(uri.getQuery() == null) {
			return new URI(uri.getScheme(), uri.getUserInfo(), uri.getHost(), uri.getPort(), uri.getPath(), "secureid="+generateSID(toSign), uri.getFragment());
		} else {
			return new URI(uri.getScheme(), uri.getUserInfo(), uri.getHost(), uri.getPort(), uri.getPath(), uri.getQuery()+"&secureid="+generateSID(toSign), uri.getFragment());
		}
		} catch (URLEncodedFormatException e) {
			Logger.error(this, "UNABLE TO ENCODE INTERNALLY GENERATED STRING: "+uri+" because URL decode failure: "+e);
			return uri;
		}
	}

	public synchronized void setNonce(byte[] nonce) {
		this.clientNonce = nonce;
	}

}
